import argparse
import glob
import os
import sys
import pandas as pd
from Bio import SeqIO
import datetime
import time
import mvip

def fetch_arguments(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="MVP_02_filter_genomad_checkv")
    parser.add_argument(
        "--input", "-i",
        help="Path to your working directory where you want to run MVP.",
    )
    parser.add_argument(
        "--metadata", "-m",
        help="Path to your metadata that you want to use to run MVP.",
    )
    parser.add_argument(
        "--sample_group",
        help="Specific sample number(s) to run the script on (can be a comma-separated list: 1,2,6 for example). By default, MVP processes all datasets listed in the metadata file one after the other.",
    )
    parser.add_argument(
        "--viral_min_genes",
        type=int,
        default=1,
        help="the minimum number of viral genes required to consider a row as a virus prediction (default = 1)",
    )
    parser.add_argument(
        "--host_viral_genes_ratio",
        type=int,
        default=1,
        help="the maximum ratio of host genes to viral genes required to consider a row as a virus prediction (default = 1",
    )

class StepCounter:
    def __init__(self):
        self.current_step = 0
        self.current_substep = 'A'

    def print_main_step(self):
        self.current_step += 1
        self.current_substep = 'A'  # Reset sub-step counter to 'A'
        return f"Step {self.current_step}"

    def print_sub_step(self):
        if self.current_substep == 'A':  # Only increment main step if starting a new sub-step series
            main_step = self.print_main_step()
        else:
            main_step = f"Step {self.current_step}"  # Use current main step without incrementing
        sub_step = self.current_substep
        self.current_substep = chr(ord(self.current_substep) + 1)
        return f"{main_step}{sub_step}"
    
    def reset(self):
        self.current_step = 0
        self.current_substep = 'A'

# Example usage:
step_counter = StepCounter()

def rename_sequence(sequence_name):
    sequence_name = sequence_name.replace('provirus_', '')
    sequence_name = sequence_name.rsplit('/', 1)[0]
    sequence_name = sequence_name.replace('|', '_').replace('-', '_').replace('/', '_').replace(':', '_')
    return sequence_name

def merge_tables(sample_name, args):
    print(f"\n\033[1m{step_counter.print_main_step()}: Merging the geNomad and CheckV output tables...\033[0m")
    # Construct the path to the quality_summary.tsv file based on the sample name
    unfiltered_virus_provirus_genomad_checkv_path = os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'MVP_01_{sample_name}_Unfiltered_Merged_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv')
    unfiltered_virus_provirus_genomad_checkv_df = pd.read_csv(unfiltered_virus_provirus_genomad_checkv_path, sep='\t')

    # Filter the merged dataframe to keep only rows predicted as 'virus'
    filtered_virus_provirus_genomad_checkv_df = unfiltered_virus_provirus_genomad_checkv_df.loc[(unfiltered_virus_provirus_genomad_checkv_df['viral_genes'] >= args["viral_min_genes"])]
    filtered_virus_provirus_genomad_checkv_df = filtered_virus_provirus_genomad_checkv_df.loc[filtered_virus_provirus_genomad_checkv_df['host_genes'] / filtered_virus_provirus_genomad_checkv_df['viral_genes'] <= args["host_viral_genes_ratio"]]

    # Save the merged Checkv and Genomad file
    filtered_virus_provirus_genomad_checkv_path = os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'MVP_02_{sample_name}_Filtered_Relaxed_Merged_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv')
    filtered_virus_provirus_genomad_checkv_df.to_csv(filtered_virus_provirus_genomad_checkv_path, sep='\t', index=False)

    return filtered_virus_provirus_genomad_checkv_df

def generate_fasta_files(virus_fasta_file, provirus_fasta_file, sample_name, args, virus_ids_to_keep, rename_sequence):
    print(f"\n\033[1m{step_counter.print_sub_step()}: Creating the unfiltered FASTA file containing virus and provirus sequences...\033[0m")
    virus_fasta_lines = []
    provirus_fasta_lines = []

    if os.path.isfile(virus_fasta_file):
        with open(virus_fasta_file, 'r') as f:
            virus_fasta_lines = f.readlines()

    if os.path.isfile(provirus_fasta_file):
        with open(provirus_fasta_file, 'r') as f:
            provirus_fasta_lines = f.readlines()

    # Concatenate the virus and provirus fasta sequences
    concatenated_fasta_lines = virus_fasta_lines + provirus_fasta_lines

    # Concatenated FASTA file
    concatenated_fasta_file = os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'MVP_02_{sample_name}_Unfiltered_Virus_Provirus_Sequences.fna')
    
    # Open the input file and create a list to store modified sequences
    with open(concatenated_fasta_file, 'w') as f:
        for line in concatenated_fasta_lines:
            if line.startswith('>'):
                contig_id = line.strip()[1:]
                # Use the rename_sequence function here
                renamed_contig_id = rename_sequence(contig_id)
                write_line = True
                f.write(f'>{renamed_contig_id}\n')
            else:
                if write_line:
                    f.write(line)

    # Open the concatenated file and read the lines
    with open(concatenated_fasta_file, 'r') as input_file:
        concatenated_fasta_lines = input_file.readlines()

    print(f"\n\033[1m{step_counter.print_sub_step()}: Creating the filtered FASTA file containing virus and provirus sequences...\033[0m")
    # Filter the concatenated file
    filtered_concatenated_fasta_file = os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'MVP_02_{sample_name}_Filtered_Relaxed_Virus_Provirus_Sequences.fna')
    
    # Open the output file for writing the filtered sequences
    with open(filtered_concatenated_fasta_file, 'w') as output_file:
        write_line = False  # Initialize the flag
        for line in concatenated_fasta_lines:
            if line.startswith('>'):
                contig_id = line.strip()[1:]
                if contig_id in virus_ids_to_keep:
                    write_line = True
                    output_file.write(line)
                else:
                    write_line = False
            else:
                if write_line:
                    output_file.write(line)

def generate_summary_report(args, sample_name, formatted_start_time, formatted_end_time, duration):
    print(f"\n\033[1m{step_counter.print_main_step()}: Generating summary report for {sample_name}...\033[0m")
   
    # Define the lines you want to add
    module_02_header = """****************************************************************************
******************               MODULE 02                ******************
****************************************************************************
"""

    # Create a dictionary to hold argument descriptions and their default values
    argument_defaults = {
        '--input': args["input"],
        '--metadata': args["metadata"],
        '--sample-group': args["sample_group"],
        '--viral-min-genes': args["viral_min_genes"],
        '--host-viral-genes-ratio': args["host_viral_genes_ratio"]}

    # Write a summary line with script arguments and their default values
    summary_line = "02_filter_genomad_checkv.py"
    for arg, default in argument_defaults.items():
        if default is not None:
            summary_line += f" {arg} {default}"

    # Specify the path for the new MVP_01_Summary_Report.txt in '01_GENOMAD' directory
    summary_report_02_path = os.path.join(args["input"], '02_CHECK_V', str(sample_name), str(sample_name) + '_MVP_02_Summary_Report.txt')

    # Write the modified content to the new summary report in '01_GENOMAD'
    with open(summary_report_02_path, 'w') as summary_report:
        summary_report.write(module_02_header + summary_line)
        summary_report.write(f"\n")
        summary_report.write(f"\nStart Time: {formatted_start_time}\n")
        summary_report.write(f"End time: {formatted_end_time}\n")
        summary_report.write(f"Running Time: {duration.total_seconds():.2f} seconds\n")

    print("Summary report for Module 02 has been created: ", summary_report_02_path)
    print(f"\nEnd Time of processing {sample_name}: {formatted_end_time}")
    print(f"Running Time of processing {sample_name}: {duration.total_seconds():.2f} seconds")

def main(args):
    # Capture start time
    mvp_start_time = datetime.datetime.now()
    mvp_formatted_start_time = mvp_start_time.strftime("%Y-%m-%d %H:%M:%S")
    print(f"\nStart Time: {mvp_formatted_start_time}") 

    # read metadata file and map FASTQ files
    metadata = pd.read_csv(args["metadata"], sep='\t')

    # Process rows based on specific sample numbers or all rows
    specific_sample_numbers = []
    if args["sample_group"]:
        specific_sample_numbers = [int(num) for num in args["sample_group"].split(',') if num.strip()]

    for row_index, (_, row) in enumerate(metadata.iterrows(), start=1):
        sample_number = row['Sample_number']
        sample_name = row['Sample']

        # Check if the current sample number is in the list of specific sample numbers
        if specific_sample_numbers and sample_number not in specific_sample_numbers:
            continue # Skip to the next iteration

        step_counter.reset()

        # Capture start time
        start_time = datetime.datetime.now()
        formatted_start_time = start_time.strftime("%Y-%m-%d %H:%M:%S")
        print(f"\nStart Time of processing {sample_name}: {formatted_start_time}") 

        filtered_virus_provirus_genomad_checkv_df = merge_tables(sample_name, args)

        virus_ids_to_keep = set(filtered_virus_provirus_genomad_checkv_df['virus_id'].tolist())

        # Concatenated FASTA sequences
        virus_fasta_file = glob.glob(os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'{sample_name}_Viruses_CheckV_Output/viruses.fna'))[0]
        provirus_fasta_file = glob.glob(os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'{sample_name}_Proviruses_CheckV_Output/viruses.fna'))[0]

        generate_fasta_files(virus_fasta_file, provirus_fasta_file, sample_name, args, virus_ids_to_keep, rename_sequence)
        
        # Capture end time
        end_time = datetime.datetime.now()
        formatted_end_time = end_time.strftime("%Y-%m-%d %H:%M:%S")

        # Calculate duration
        duration = end_time - start_time

        # Define the additional lines you want to add to the summary report
        generate_summary_report(args, sample_name, formatted_start_time, formatted_end_time, duration)

    # Capture end time
    mvp_end_time = datetime.datetime.now()
    mvp_formatted_end_time = mvp_end_time.strftime("%Y-%m-%d %H:%M:%S")

    # Calculate duration
    mvp_duration = mvp_end_time - mvp_start_time

    message1 = "\033[1mModule 02 finished: output summary files merged and filtered successfully!\033[0m\n"
    message2 = "Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv and viruses_proviruses.fna files have been generated for each sample in their respective directories in 02_CHECK_V."
    message3 = "viruses_proviruses.fna files will be used to run Module 03 (clustering)."
    message4 = "\n\033[1mYou can now proceed to the next step of the MVP script: Module 03!\033[0m"
    line_of_stars = '*' * len(message2)
    print()
    print(line_of_stars)
    print(message1)
    print(f"Start Time: {mvp_formatted_start_time}") 
    print(f"End Time: {mvp_formatted_end_time}")
    print(f"Running Time: {mvp_duration.total_seconds():.2f} seconds\n")
    print(message2)
    print(message3)
    print(message4)
    print(line_of_stars)
    print()

    print("Please don't forget to cite MVP.")


#! /usr/bin/env python

import Bio.SeqIO, time, gzip
from operator import itemgetter
import numpy as np, argparse


def parse_blast(handle):
    for line in handle:
        r = line.split()
        yield {
            "qname": r[0],
            "tname": r[1],
            "pid": float(r[2]),
            "len": float(r[3]),
            "qcoords": sorted([int(r[6]), int(r[7])]),
            "tcoords": sorted([int(r[8]), int(r[9])]),
            "qlen": float(r[-2]),
            "tlen": float(r[-1]),
            "evalue": float(r[-4]),
        }


def yield_alignment_blocks(handle):
    # init block with 1st record
    key, alns = None, None
    for aln in parse_blast(handle):
        key = (aln["qname"], aln["tname"])
        alns = [aln]
        break
    # loop over remaining records
    for aln in parse_blast(handle):
        # extend block
        if (aln["qname"], aln["tname"]) == key:
            alns.append(aln)
        # yield block and start new one
        else:
            yield alns
            key = (aln["qname"], aln["tname"])
            alns = [aln]
    yield alns


def prune_alns(alns, min_length=0, min_evalue=1e-3):
    # remove alignments with < min_length or > min_evalue
    # discard alignments after the entire query length has been covered
    keep = []
    cur_aln = 0
    qry_len = alns[0]["qlen"]
    for aln in alns:
        qcoords = aln["qcoords"]
        aln_len = max(qcoords) - min(qcoords) + 1
        if aln_len < min_length or aln["evalue"] > min_evalue:
            continue
        if cur_aln >= qry_len or aln_len + cur_aln >= 1.10 * qry_len:
            break
        keep.append(aln)
        cur_aln += aln_len
    return keep


def compute_ani(alns):
    return round(
        sum(a["len"] * a["pid"] for a in alns) / sum(a["len"] for a in alns), 2
    )


def compute_cov(alns):

    # merge qcoords
    coords = sorted([a["qcoords"] for a in alns])
    nr_coords = [coords[0]]
    for start, stop in coords[1:]:

        # overlapping, update start coord
        if start <= (nr_coords[-1][1] + 1):
            nr_coords[-1][1] = max(nr_coords[-1][1], stop)

        # non-overlapping, append to list
        else:
            nr_coords.append([start, stop])

    # compute qry_cov
    alen = sum([stop - start + 1 for start, stop in nr_coords])
    qcov = round(100.0 * alen / alns[0]["qlen"], 2)

    # merge tcoords
    coords = sorted([a["tcoords"] for a in alns])
    nr_coords = [coords[0]]
    for start, stop in coords[1:]:
        # overlapping, update start coord
        if start <= (nr_coords[-1][1] + 1):
            nr_coords[-1][1] = max(nr_coords[-1][1], stop)
        # non-overlapping, append to list
        else:
            nr_coords.append([start, stop])

    # compute qry_cov
    alen = sum([stop - start + 1 for start, stop in nr_coords])
    tcov = round(100.0 * alen / alns[0]["tlen"], 2)

    return qcov, tcov


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        dest="input",
        type=str,
        required=True,
        metavar="PATH",
        help="path to blastn input file (format: 'std 6 qlen slen')",
    )
    parser.add_argument(
        "-o",
        dest="output",
        type=str,
        required=True,
        metavar="PATH",
        help="path to ani file",
    )
    parser.add_argument(
        "-l",
        dest="length",
        type=int,
        metavar="INT",
        help="minimum alignment length to keep",
    )
    return vars(parser.parse_args())


if __name__ == "__main__":
    args = parse_arguments()
    out = (
        gzip.open(args["output"], "w")
        if args["output"].split(".")[-1] == "gz"
        else open(args["output"], "w")
    )
    fields = ["qname", "tname", "num_alns", "pid", "qcov", "tcov"]
    out.write("\t".join(fields) + "\n")
    input = (
        gzip.open(args["input"])
        if args["input"].split(".")[-1] == "gz"
        else open(args["input"])
    )
    for alns in yield_alignment_blocks(input):
        alns = prune_alns(alns)
        if len(alns) == 0:
            continue
        qname, tname = alns[0]["qname"], alns[0]["tname"]
        ani = compute_ani(alns)
        qcov, tcov = compute_cov(alns)
        row = [qname, tname, len(alns), ani, qcov, tcov]
        out.write("\t".join([str(_) for _ in row]) + "\n")



import argparse
import os
import subprocess
import sys
from os.path import getsize
import pandas as pd
from functools import reduce
from Bio import SeqIO
from tqdm import tqdm
import datetime
import time
import mvip

def fetch_arguments(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="MVP_00_set_up_MVP")
    parser.add_argument(
        "--input", "-i",
        type=str,
        help="Path to the working directory where MVP will be run.",
    )
    parser.add_argument(
        "--metadata", "-m",
        type=str,
        help="Path to the metadata that will be use to run MVP.",
    )
    parser.add_argument(
        "--skip_install_databases",
        action="store_true",
        default=False,
        help="Install geNomad and CheckV databases in the respective repositories. (default = False).",
    )
    parser.add_argument(
        '--genomad_db_path', 
        type=str,
        help='Path to the directory where geNomad database will be installed.'
        )
    parser.add_argument(
        '--checkv_db_path', 
        type=str,
        help='Path to the directory where CheckV database will be installed.'
        )
    parser.add_argument(
        "--skip_check_errors",
        action="store_true",
        default=False,
        help="Skip to run sequence data error checking. (default = False).",
    )
    
# Global variable to track the current step number
current_step = 0
# Dictionary to track sub-step counts for each function
sub_step_counts = {}

def print_step(func_name):
    global current_step
    # If it's the first call to this function, increment the step number
    if func_name not in sub_step_counts:
        current_step += 1
        sub_step_counts[func_name] = 1
        return f"Step {current_step}"  # Return without a letter for the first call
    else:
        # Increment the sub-step count for subsequent calls within the same function
        sub_step_counts[func_name] += 1
        sub_step_letter = chr(64 + sub_step_counts[func_name])  # 'A' is 65 in ASCII
        return f"Step {current_step}{sub_step_letter}"


def convert_metadata_to_unix(args):
    print(f"\n\033[1m{print_step(convert_metadata_to_unix)}: Converting the metadata file...\033[0m")

    subprocess.run(['dos2unix', args["metadata"]])

def check_metadata(args):
    print(f"\n\033[1m{print_step(check_metadata)}: Checking for missing, incorrectly named mandatory columns or extra columns in the metadata file...\033[0m")

    # Read the metadata file
    metadata = pd.read_csv(args["metadata"], sep='\t')

    # Define the mandatory and optional columns
    mandatory_columns = ['Sample_number', 'Sample', 'Assembly_Path', 'Read_Path']
    optional_columns = ['Variable']

    missing_columns = [col for col in mandatory_columns if col not in metadata.columns]
    if missing_columns:
        print(f"Error: Missing mandatory column(s) in metadata: {', '.join(missing_columns)}")
        sys.exit(1)

    # Check for extra columns
    extra_columns = [col for col in metadata.columns if col not in mandatory_columns + optional_columns]
    if extra_columns:
        print(f"Error: Extra column(s) found in metadata: {', '.join(extra_columns)}")
        sys.exit(1)

    print("The metadata looks fine. Proceeding to the next steps.")

def check_paths(args):
    print(f"\n\033[1m{print_step(check_paths)}: Checking if the assembly or read file paths provided in the metadata are correct...\033[0m")

    metadata = pd.read_csv(args["metadata"], sep='\t')
    nonexistent_assembly_paths = metadata.loc[~metadata['Assembly_Path'].apply(lambda path: path == '' or (isinstance(path, str) and os.path.exists(path))), 'Assembly_Path']
    nonexistent_read_paths = metadata.loc[~metadata['Read_Path'].apply(lambda path: path == '' or (isinstance(path, str) and os.path.exists(path))), 'Read_Path']

    if not nonexistent_assembly_paths.empty or not nonexistent_read_paths.empty:
        if not nonexistent_assembly_paths.empty:
            print(f"Error: The following Assembly_Path values correspond to non-existent files: {', '.join(nonexistent_assembly_paths)}")
        if not nonexistent_read_paths.empty:
            nonexistent_read_paths_str = nonexistent_read_paths.astype(str)
            print(f"Error: The following Read_Path values correspond to non-existent files: {', '.join(nonexistent_read_paths_str)}")
        sys.exit(1)
    
    print("The assembly and read file paths are correct. Proceeding to the next steps.")

def check_sequence_data(args):
    if args["skip_check_errors"]:
        print(f"\n\033[1m{print_step(check_sequence_data)}: Skipping verifying that the input assembly files contain only valid nucleotide characters and do not have duplicate headers...\033[0m")
        return
    
    print(f"\n\033[1m{print_step(check_sequence_data)}: Verifying that the input assembly files contain only valid nucleotide characters and does not have duplicate headers...\033[0m")
    metadata = pd.read_csv(args["metadata"], sep='\t')
    for row_index, (_, row) in enumerate(metadata.iterrows(), start=1):
        input_assembly_file = row['Assembly_Path']
        
        with open(input_assembly_file, "r") as input_fasta:
            headers_seen = set()
            invalid_chars_found = False
            duplicate_headers_found = False
            
            for record in SeqIO.parse(input_fasta, "fasta"):
                header = record.id
                if any(base not in "ACTGNactgn" for base in record.seq):
                    print(f"Error: Invalid nucleotide characters found in {input_assembly_file} for sequence {record.id}\n")
                    invalid_chars_found = True
                if header in headers_seen:
                    print(f"Error: Duplicate header found in {input_assembly_file}: {header}\n")
                    duplicate_headers_found = True
                headers_seen.add(header)
            
            if invalid_chars_found or duplicate_headers_found:
                sys.exit(1)

    print("The input assembly files look fine. Proceeding to the next steps.")

def create_output_folders(args):
    print(f"\n\033[1m{print_step(create_output_folders)}: Creating MVP directories where outputs will be stored...\033[0m")

    output_folders = ['00_DATABASES', '01_GENOMAD', '02_CHECK_V', '03_CLUSTERING', '04_READ_MAPPING', '05_VOTU_TABLES', '06_FUNCTIONAL_ANNOTATION']
    for folder_name in output_folders:
        output_folder = os.path.join(args["input"], folder_name)
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)

def install_databases(args):
    if args["skip_install_databases"]:
        genomad_db_path = None
        checkv_db_path = None
        print(f"\n\033[1m{print_step(install_databases)}: Skipping to install databases because --install_databases is False...\033[0m")
    else:
        database_path = os.path.join(args["input"], '00_DATABASES')
        os.makedirs(database_path, exist_ok=True)

        # Install geNomad database
        if args['genomad_db_path']:
            genomad_db_path = os.path.join(args['genomad_db_path'], 'genomad_db')
        else:
            genomad_db_path = os.path.join(database_path, 'genomad_db')
        
        if os.path.exists(genomad_db_path):
            print(f"\n\033[1m{print_step(install_databases)}: Skipping to install geNomad database because {genomad_db_path} already exists...\033[0m")
        else:
            print(f"\n\033[1m{print_step(install_databases)}: Creating geNomad database...\033[0m")
            subprocess.run(['genomad', 'download-database', database_path])

        # Install CheckV database
        if args['checkv_db_path']:
            checkv_db_path = os.path.join(args['checkv_db_path'], 'checkv-db-v1.5')
        else:
            checkv_db_path = os.path.join(database_path, 'checkv-db-v1.5')

        if os.path.exists(checkv_db_path):
            print(f"\n\033[1m{print_step(install_databases)}: Skipping to install Check_V database because {checkv_db_path} already exists...\033[0m")
        else:
            print(f"\n\033[1m{print_step(install_databases)}: Creating CheckV database...\033[0m")
            subprocess.run(['checkv', 'download_database', database_path])

        additional_dbs = ['PhrogDB_v14', 'RdRP_DB', 'Pfam_A_DB', 'dbAPIS']
        for db in additional_dbs:
            db_path = os.path.join(database_path, db)
            if os.path.exists(db_path) and os.listdir(db_path):
                print(f"\n\033[1m{print_step(install_databases)}: Skipping install of {db} because it already exists...\033[0m")
            else:
                print(f"\n\033[1m{print_step(install_databases)}: Downloading {db}...\033[0m")
                subprocess.run(['wget', '-r', '-np', '-nH', '--cut-dirs=3', '--reject="index.html*"', f'https://portal.nersc.gov/cfs/m342/MViP/{db}/', '-P', database_path])

    return genomad_db_path, checkv_db_path

def generate_summary_report(args, genomad_db_path, checkv_db_path, formatted_start_time, formatted_end_time, duration):
    print(f"\n\033[1m{print_step(generate_summary_report)}: Generating summary report for Module 00...\033[0m")

    # Create a dictionary to hold argument descriptions and their default values
    argument_defaults = {
        '--input': args["input"],
        '--metadata': args["metadata"],
        '--skip_check_errors': args["skip_check_errors"],
        '--skip_install_databases': args["skip_install_databases"],
        '--genomad_db_path': genomad_db_path,
        '--checkv_db_path': checkv_db_path
    }

    # Create the module header
    module_header = """
****************************************************************************
*********                          MODULE 00                       *********
****************************************************************************
"""

    # Write a summary line with script arguments and their default values
    summary_line = "00_set_up_mvp.py"
    for arg, default in argument_defaults.items():
        if default is not None:
            summary_line += f" {arg} {default}"

    # Create a summary report file
    summary_report_path = os.path.join(args["input"], 'MVP_00_Summary_Report.txt')

    with open(summary_report_path, 'w') as summary_report:
        # Write the module header
        summary_report.write(module_header.strip() + '\n')
        summary_report.write(summary_line + '\n\n')
        summary_report.write(f"Start Time: {formatted_start_time}\n")
        summary_report.write(f"End Time: {formatted_end_time}\n")
        summary_report.write(f"Running Time: {duration.total_seconds():.2f} seconds\n\n")
        summary_report.write("Tools and versions:\n")

        # List of tools
        tools = [
            "tqdm", "genomad", "pandas", "checkv", "bowtie2", "minimap2", 
            "samtools", "seqtk", "scikit-learn", "numpy", "parallel", 
            "coverm", "python", "dos2unix", "mmseqs2", "mafft", "trimal", 
            "fasttree", "vRhyme", "taxopy", "table2asn"
        ]

        # For each tool, find it in conda list and get its version
        for tool in tools:
            try:
                result = subprocess.run(
                    ["conda", "list", tool], capture_output=True, text=True, check=True
                )
                lines = result.stdout.splitlines()
                if len(lines) > 3:
                    version_line = lines[3]  # Assuming the version info is in the 4th line
                    version = version_line.split()[1]
                    summary_report.write(f"{tool} {version}\n")
                else:
                    summary_report.write(f"{tool} version not found\n")
            except subprocess.CalledProcessError:
                summary_report.write(f"{tool} version not found\n")

        # Add section for databases and versions
        summary_report.write("\nDatabases and versions:\n")
        if genomad_db_path:
            genomad_db_name = os.path.basename(genomad_db_path)
            # Read version from version.txt
            version_file_path = os.path.join(genomad_db_path, 'version.txt')
            if os.path.exists(version_file_path):
                with open(version_file_path, 'r') as version_file:
                    genomad_version = version_file.read().strip()
                summary_report.write(f"{genomad_db_name} {genomad_version}\n")
            else:
                summary_report.write(f"{genomad_db_name}: version file not found\n")
        else:
            summary_report.write("geNomad database: not installed\n")
        
        if checkv_db_path:
            checkv_db_name = os.path.basename(checkv_db_path)
            # Extract version number from checkv_db_path
            checkv_version = checkv_db_name.split('-')[-1].replace('v', '')
            summary_report.write(f"{checkv_db_name} {checkv_version}\n")
        else:
            summary_report.write("CheckV database: not installed\n")

    print("Summary report for Module 00 has been created:", summary_report_path)

def main(args):
    # Capture start time
    start_time = datetime.datetime.now()
    formatted_start_time = start_time.strftime("%Y-%m-%d %H:%M:%S") 
    print(f"\nStart Time: {formatted_start_time}")

    # Usage of run_step for the dos2unix step
    convert_metadata_to_unix(args)

    check_metadata(args)

    # Usage of run_step for the check_paths step
    check_paths(args)

    # Usage of run_step for the check_sequence_data step
    check_sequence_data(args)

    create_output_folders(args)

    genomad_db_path, checkv_db_path = install_databases(args)

    # Capture end time
    end_time = datetime.datetime.now()
    formatted_end_time = end_time.strftime("%Y-%m-%d %H:%M:%S")

    # Calculate duration
    duration = end_time - start_time

    # Add this function to your script
    generate_summary_report(args, genomad_db_path, checkv_db_path, formatted_start_time, formatted_end_time, duration)

    message1 = "\033[1mModule 00 finished: you are all set to use MVP!\n\033[0m"
    message2 = "Your metadata looks fine, as well as your read and assembly files."
    message3 = "All the directories needed for MVP have been generated in your working directory."
    if not args["skip_install_databases"]:
        message4 = "geNomad and CheckV databases have been installed in their respective directories."
    message5 = "\n\033[1mYou can now proceed to the next step of the MVP script: Module 01!\033[0m"
    line_of_stars = '*' * len(message3)

    print()
    print(line_of_stars)
    print(message1)
    print(f"Start Time: {formatted_start_time}")
    print(f"End Time: {formatted_end_time}")
    print(f"Running Time: {duration.total_seconds():.2f} seconds\n")
    print(message2)
    print(message3)
    if not args["skip_install_databases"]:
        print(message4)
    print(message5)
    print(line_of_stars)
    print()

    print("Please don't forget to cite MVP and the following softwares used in this module:")
    print("- Camargo, A. P., Roux, S., Schulz, F., Babinski, M., Xu, Y., Hu, B., Chain, P. S. G., Nayfach, S., & Kyrpides, N. C. You can move, but you can’t hide: identification of mobile genetic elements with geNomad. bioRxiv (2023), DOI: 10.1101/2020.11.01.361691")
    print("\n- Nayfach, S., Camargo, A.P., Schulz, F. et al. CheckV assesses the quality and completeness of metagenome-assembled viral genomes. Nat Biotechnol 39, 578–585 (2021). https://doi.org/10.1038/s41587-020-00774-7\n")
